      program cfd

      implicit none

      include 'mpif.h'

      integer m, n

      integer ln

      double precision, allocatable ::  psi(:,:), tmp(:,:), vels(:,:,:)

      integer, allocatable :: rgb(:,:,:)

      integer scalefactor
      integer numiter
      character*10 temp_char10

!
!  Some auxiliary parameters and variables
!

      integer b, h, w

      integer iter, i, j

      double precision tstart, tstop, time, hue
      double precision, allocatable ::  modvsq(:,:)

!
!  Variables needed for parallelisation
!

      integer comm, commline, rank, size, ierr
      integer shift, dim, uprank, dnrank
      logical periodic, reorder

      integer jstart, jstop

!
!  Parallel initialisation
!

      comm = MPI_COMM_WORLD

      call mpi_init(ierr)

      call mpi_comm_rank(comm, rank, ierr)
      call mpi_comm_size(comm, size, ierr)

!
!  Read in parameters
!

!       scalefactor = 1
!       numiter = 1000

        if (rank .eq. 0) then 
           call getarg(1,temp_char10)
           read(temp_char10,*) scalefactor
           call getarg(2,temp_char10)
           read(temp_char10,*) numiter
 
           print*, 'scalefactor, number of iterations = '
           print*,  scalefactor, numiter
        end if

!broadcast scalefactor and numiter to all the other processors

	call mpi_bcast(scalefactor,1,MPI_INTEGER,0,comm,ierr)
        call mpi_bcast(numiter,1,MPI_INTEGER,0,comm,ierr)

! calculate b, h & w and m & n
        
        b = 5*scalefactor 
        h = 15*scalefactor
        w = 5*scalefactor 
        m = 48*scalefactor
        n = 48*scalefactor
	ln = n/size

!allocate 
	allocate(psi(0:m+1, 0:ln+1))
	allocate(tmp(0:m+1, 0:ln+1))
        allocate(vels(2, m, ln))
	allocate(rgb(3,m,ln))
	allocate(modvsq(m,ln))

      if (size*ln .ne. n) then

        if (rank .eq. 0) then
          write(*,*) 'ERROR:'
          write(*,*) 'n = ', n, ' does not divide onto ', size, ' PEs'
        end if

        call mpi_finalize(ierr)

        stop

      end if

      if (rank .eq. 0) then
        write(*,10) m, n, size
10      format(1x, 'Running CFD on ', i4, ' x ', i4, ' grid using ',
     &          i3, ' processors')
        write(*,*)
      end if


      jstart = ln*rank + 1
      jstop  = jstart + ln - 1

!
!  Compute ranks of nearest neighbours
!

      periodic = .FALSE.
      reorder  = .TRUE.
      
      call mpi_cart_create(comm, 1, size, periodic, reorder,
     &                     commline, ierr)

      dim = 0
      shift = 1

      call mpi_cart_shift(commline, dim, shift, dnrank, uprank, ierr)

!
!  Zero the psi array
!

      do i = 0, m+1
        do j = 0, ln+1

          psi(i,j) = 0.0
 
        end do
      end do

!
!  Set the boundary conditions on the bottom edge
!

      if (rank .eq. 0) then

        do i = b+1, b+w-1
          psi(i, 0) = float(i-b)
        end do

        do i = b+w, m
           psi(i, 0) = float(w)
        end do

        
      end if

!
!  Set the boundary conditions on the right hand side
!

      do j = 1, h

        if (j .ge. jstart .and. j .le. jstop) then
          psi(m+1,j-jstart+1) = float(w)
        end if

      end do

      do j = h+1, h+w-1

        if (j .ge. jstart .and. j .le. jstop) then
          psi(m+1,j-jstart+1) = float(w-j+h)
        end if

      end do

!
!  Begin iterative Jacobi loop
!

      if (rank .eq. 0) then
        write(*,*) 'Starting main loop ...'
      end if

      call mpi_barrier(commline, ierr)

      tstart = mpi_wtime()

      do iter = 1, numiter

!
!  Do a boundary swap
!

      call haloswap(psi, m, ln, uprank, dnrank, commline)

!
!  Compute the new psi based on the old one
!

        do i = 1, m
          do j = 1, ln

            tmp(i,j) = (psi(i+1,j)+psi(i-1,j)+psi(i,j+1)+psi(i,j-1))/4.0

          end do
        end do

!
!  Update psi
!

        do i = 1, m
          do j = 1, ln

            psi(i,j) = tmp(i,j)

          end do
        end do

!
!  End iterative Jacobi loop
!

      end do

      call mpi_barrier(commline, ierr)

      tstop = mpi_wtime()

      time = (tstop-tstart)/float(numiter)

      if (rank .eq. 0) then

        write(*,*) '... finished'
        write(*,*)
        write(*,20) numiter, time, tstop-tstart
20      format(1x, 'Ran for ', i5, ' iterations, ',
     &          g12.6, ' seconds per iteration with total time ', g12.6)
 
        write(*,*) 
        write(*,*) 'Writing output file ...'
      end if

!
!  Output result
!

!
!  Another boundary swap since v depends on nonlocal psi
!

      call haloswap(psi, m, ln, uprank, dnrank, commline)

      do j = 1, ln
        do i = 1, m

          modvsq(i,j) = (  ((psi(i,j+1) - psi(i,j-1))/2.0)**2
     &                   + ((psi(i-1,j) - psi(i+1,j))/2.0)**2 )

          hue = modvsq(i,j)**0.4

          call hue2rgb(hue, rgb(1,i,j), rgb(2,i,j), rgb(3,i,j))

        end do
      end do

      call pdump(rgb, m, ln, size, rank, comm)

!
! Dump psi, modvsq, and u to AVS/Express file
!
! [AP, 25/11/2013] - Have disabled this for the time being
!                    rely on .ppm file for visualisation
!
      call calc_vels(psi,vels,m,ln)
!      call write_psi(psi, modvsq, vels, m, ln, size, rank, comm,
!     $     "cfd-output.fld")

      deallocate(psi, tmp, vels, rgb, modvsq)

      call mpi_finalize(ierr)

      if (rank .eq. 0) then
        write(*,*) ' ... finished'
        write(*,*)
        write(*,*) 'CFD completed'
      end if

      end


      subroutine haloswap(x, m, n, uprank, dnrank, comm)

      implicit none

      include 'mpif.h'

      integer m, n, uprank, dnrank, comm

      double precision x(0:m+1,0:n+1)

      integer tag, request, ierr
      parameter (tag = 1)

      integer status(MPI_STATUS_SIZE)

!
!  Send upper boundaries and receive lower ones
!

      call mpi_isend(x(1,n), m, MPI_DOUBLE_PRECISION, uprank,
     &                tag, comm, request, ierr)

      call mpi_recv(x(1,0), m, MPI_DOUBLE_PRECISION, dnrank,
     &              tag, comm, status, ierr)

      call mpi_wait(request, status, ierr)

!
!  Send lower boundaries and receive upper ones
!

      call mpi_isend(x(1,1), m, MPI_DOUBLE_PRECISION, dnrank,
     &                tag, comm, request, ierr)

      call mpi_recv(x(1,n+1), m, MPI_DOUBLE_PRECISION, uprank,
     &              tag, comm, status, ierr)

      call mpi_wait(request, status, ierr)

      return
      end


      subroutine pdump(rgb, m, n, size, rank, comm)

      implicit none

      include 'mpif.h'

      integer comm, size, rank, m, n

      integer rgb(3,m,n)

      integer irank, i, j, k, iounit, tag, ierr

      parameter (tag = 1)
      parameter (iounit = 10)

      integer status(MPI_STATUS_SIZE)

      if (rank .eq. 0) then

        open(unit=iounit, file='output.ppm', form='formatted')

        write(unit=iounit, fmt=10) m, n*size, 255
10      format('P3',/,i4,2x,i4,/,i3)

        do irank = 0, size-1

          if (irank .ne. 0) then

            call mpi_recv(rgb, 3*m*n, MPI_INTEGER, irank, tag,
     &                    comm, status, ierr)

          end if

          write(unit=iounit, fmt=20) (((rgb(k,i,j),k=1,3),i=1,m),j=1,n)
20        format(18(i4))

        end do

        close(unit=iounit)

      else

        call mpi_ssend(rgb, 3*m*n, MPI_INTEGER, 0, tag, comm, ierr)

      end if

      return
      end


      subroutine hue2rgb(hue, r, g, b)

      implicit none

      double precision hue, colfunc

      integer r, g, b, rgbmax
      parameter (rgbmax = 255)

      r = rgbmax*colfunc(hue-1.0)
      g = rgbmax*colfunc(hue-0.5)
      b = rgbmax*colfunc(hue    )

      return
      end


      double precision function colfunc(x)

      implicit none

      double precision x
      double precision absx, x1, x2, val

      parameter (x1 = 0.2, x2 = 0.5)

      absx = abs(x)

      if (absx .gt. x2) then
        val = 0.0
      else if (absx .lt. x1) then
        val = 1.0
      else
        val = 1.0 - ((absx-x1)/(x2-x1))**2
      end if

      colfunc = val

      return
      end


      subroutine write_psi(psi, modvsq, vels, m, n, size, rank, comm,
     $     filename)

      implicit none

      include 'mpif.h'
      
      double precision modvsq(m,n), psi(0:m+1,0:n+1), vels(2,m,n)

      integer m, n, size, rank, comm 

      character(len=*) filename

      integer irank, i, j, k, iounit, tag, ierr

      parameter (tag = 2)
      parameter (iounit = 10)

      integer status(MPI_STATUS_SIZE), DT_psi_nohalo

!     Write AVS header on PE 0
      if (rank .eq. 0) then 

!     Use Fortran's primitive I/O functionalities...
         open(unit=iounit, file=filename)
 
         write(unit=iounit,fmt=10)
 10      format('# AVS')
         write(unit=iounit,fmt=20)
 20      format('#')         
         write(unit=iounit,fmt=30)
     $        'ndim=  ',2,  '   # number of dimensions in the field'
         write(unit=iounit,fmt=30)
     $        'dim1=  ', m, '   # dimension of axis 1'
         write(unit=iounit,fmt=30)
     $        'dim2=  ', n*size, '   # dimension of axis 2'
         write(unit=iounit,fmt=30)
     $        'nspace=  ',2,'   # physical coordinates per point'
         write(unit=iounit,fmt=30)
     $        'veclen=  ',4,'   # components at each point'
 30      format(A7,I4,A45)
         write(unit=iounit,fmt=40)
     $        'data=double                                   # datatype'
         write(unit=iounit,fmt=40)
     $        'field=uniform                               # field type'
         write(unit=iounit,fmt=40)
     $        'label=modvsq                     # label for component 1'
         write(unit=iounit,fmt=40)
     $        'label=psi                        # label for component 2'
         write(unit=iounit,fmt=40)
     $        'label=Vx                         # label for component 3'
         write(unit=iounit,fmt=40)
     $        'label=Vy                         # label for component 4'
 40      format(A56)
         write(unit=iounit,fmt=20)
         write(unit=iounit,fmt=50) 1, 0, 0, 1
         write(unit=iounit,fmt=50) 2, n*m*size, 0, 1
         write(unit=iounit,fmt=50) 3, 2*n*m*size, 0, 2
         write(unit=iounit,fmt=50) 4, 2*n*m*size+1, 0, 2
 50      format('variable ',I2,' file=output.dat filetype=ASCII',
     $        ' skip=',I10,' offset=',I2,' stride=',I2)
         close(unit=iounit)

!     ... start writing the data (modvsq)
!     AVS/Express had difficulties with Fortran's "binary format". use ASCII

         open(unit=iounit, file='output.dat',form='formatted')

         do irank = 0, size-1
            
            if (irank .ne. 0) then
                              
               call mpi_recv(modvsq, m*n, MPI_DOUBLE_PRECISION, irank,
     $              tag, comm, status, ierr)
               
            end if

            write(unit=iounit,fmt=60) ((modvsq(i,j),i=1,m),j=1,n)
 60         format(G12.6)
         end do

      else
         
         call mpi_ssend(modvsq, m*n, MPI_DOUBLE_PRECISION, 0, tag, comm,
     $        ierr)
         
      end if

!     ... then psi
      if (rank .eq. 0) then

         do irank = 0, size-1
            
            if (irank .ne. 0) then
               
               call mpi_recv(psi, (m+2)*(n+2), MPI_DOUBLE_PRECISION,
     $              irank, tag, comm, status, ierr)
               
            end if
            
            write(unit=iounit,fmt=60) ((psi(i,j),i=1,m),j=1,n)

         end do

      else

         call mpi_ssend(psi, (m+2)*(n+2), MPI_DOUBLE_PRECISION, 0, tag,
     $        comm, ierr)
         
      end if

!     ... and finally complete data dump with the velocities
      if (rank .eq. 0) then

         do irank = 0, size-1
            
            if (irank .ne. 0) then
               
               call mpi_recv(vels, 2*m*n, MPI_DOUBLE_PRECISION,
     $              irank, tag, comm, status, ierr)
               
            end if

            write(unit=iounit, fmt=60) (((vels(k,i,j),k=1,2),i=1,m),
     $           j=1,n)

         end do

         close(unit=iounit)

      else

         call mpi_ssend(vels, 2*m*n, MPI_DOUBLE_PRECISION, 0, tag,
     $        comm, ierr)
         
      end if

      return
      end


      subroutine calc_vels(psi, vel, m, n)

      implicit none

      double precision vel(2,m,n), psi(0:m+1,0:n+1)

      integer m, n, i, j

      do i = 1, m
         do j = 1, n
            
            vel(1,i,j) = (psi(i,j+1)-psi(i,j-1)) / 2.0
            vel(2,i,j) = (psi(i+1,j)-psi(i-1,j)) / 2.0
            
         end do
      end do

      return
      end
